/*
    Copyright  1995-2003, The AROS Development Team. All rights reserved.
    $Id: DiskSpeed.c,v 1.8 2013/05/09 18:57:23 itix Exp $

    Desc: Disk subsystem Benchmarking Utility..
    Lang: English.

    Based on DiskSpeed v4.2 by Michael Sinz
*/

/*********************************************************************************************/

#ifdef __MORPHOS__
#define D(x)
#define bug
#endif

#include <string.h>

#include <exec/execbase.h>
#include <exec/io.h>
#include <exec/ports.h>
#include <exec/system.h>
#include <devices/timer.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <devices/trackdisk.h>
#include <libraries/mui.h>
#include <workbench/workbench.h>
#include <workbench/startup.h>

#include <proto/alib.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/icon.h>
#include <proto/intuition.h>
#include <proto/locale.h>
#include <proto/muimaster.h>
#include <proto/threadpool.h>
#include <proto/timer.h>
#include <proto/utility.h>

#include "DiskSpeed.h"
#include "Main.h"

#include "appclass.h"
#include "methodstack.h"
#include "mui.h"

/*********************************************************************************************/

struct DiskSpeedMUI_Intern {
    Object          *disk_app;                  /* our application itself..     */
    Object          *disk_main_win;             /* The Main Benchmark Window    */
    Object          *disk_config_win;           /* The Configuration Window     */

    Object          *devicestr;
    Object          *commentstr;

    Object          *resultlv;

	APTR btBegin;
	APTR btAbort;
	APTR btSaveResults;
};

struct DiskSpeedConfig_Intern {
    #ifndef SCSI_SPEED
    Object          *dir_checkmark;             /* Benchmark Configuration Option Objects */
    #endif
    Object          *seek_checkmark;
    Object          *mintime;

    Object          *buffer1;
    Object          *buffer2;
    Object          *buffer3;
    Object          *buffer4;

    Object          *long_checkmark;
    Object          *word_checkmark;
    Object          *byte_checkmark;
};

static struct DiskSpeedMUI_Intern diskspeedmui_intern;
static struct DiskSpeedConfig_Intern diskspeedmui_config;

enum
{
	RETURNID_BEGIN = 1,
	RETURNID_STOP,
	RETURNID_SAVE,
	RETURNID_BENCHMARK_FINISHED,
};

/*********************************************************************************************/

#include <limits.h>
#include <stdio.h>

#undef DEBUG

/*********************************************************************************************/

#ifdef	SCSI_SPEED
#   include	"ScsiSpeed_rev.h"
#else
#   include	"DiskSpeed_rev.h"
#endif	/* SCSI_SPEED */

/*********************************************************************************************/

#define VERS			        APPNAME " 5.0"
#define	VSTRING                 APPNAME " 5.0 (" DATE ")\n\r"
#define STR_SEPERATOR               "------------------------------------------------------------------"

static const TEXT COPYRIGHT[] = VERS " Copyright  1995-2004 The AROS Development Team, 2012-2013 The MorphOS Team " VSTRING;
static const TEXT VERSTAG[]   = "\0$VER: " APPNAME " 5.0 (" DATE ")";

#ifdef	SCSI_SPEED
#   define	MSG_BYTES_READ              MSG_BYTES_READ_S
    const char	RESULTS_FILE[]=         APPNAME ".Results";
#else
#   define	MSG_BYTES_READ              MSG_BYTES_READ_N
    const char	FILE_STRING[]=          "%04lx " APPNAME " Test File ";
    const char	TEST_DIR[]=             " " APPNAME " Test Directory ";
    const char	RESULTS_FILE[]=         APPNAME ".Results";
#endif	/* SCSI_SPEED */

/*********************************************************************************************/
/*
    Timer related stuff...
*/

struct SpeedTimer
{
	struct MsgPort      tmPort;
	struct timerequest  tmReq;
	struct Interrupt    tmInt;

	BOOL                TimeIsUp;
	BOOL                Open;

	UQUAD               StartTime;
	DOUBLE              FinalTime;   /* Final time in seconds */
};

/*********************************************************************************************/
/*
    Our Disk/ScsiSpeed global structure...
*/

struct DiskSpeed
{
#ifdef	SCSI_SPEED

    struct	IOExtTD	                DiskIO;                 /* Disk IO request */

#endif	/* SCSI_SPEED */

	BPTR                            Output;                 /* The output file handle! */
    struct	Process                 *Me;                    /* Pointer to my process... */
    struct	SpeedTimer              *timer;                 /* Pointer to a timer structure */

/* */
	ULONG                           Min_Time;               /* Minimum tmVal in seconds for a test */
	DOUBLE                          CPU_Usage;
	ULONG                           CPU_Total;              /* Sum of CPU availability */
	ULONG                           CPU_Count;              /* Count of CPU availability */

/* Testing parameters */
	unsigned short                  Test_DIR;               /* Set to test directory stuff */
	unsigned short                  Test_SEEK;              /* Set to test SEEK/READ */

	ULONG                           Align_Types;            /* Set bits of alignment types... */
	ULONG                           Mem_TYPES;              /* Set memory type flags to test... */
	ULONG                           Test_Size[4];           /* The four test sizes... */

/* */
	char                            Device[256];            /* Device name under test... */
	char                            Comments[256];          /* Comments string gadget... */

/* */
	CONST_STRPTR                    CPU_Type;               /* Processor name in this string (plus NULL) */
	char                            Exec_Ver[14];           /* Version of Exec */

/* */
	char                            tmp1[128];              /* Temporary buffer space... */
	char                            tmp2[128];              /* Temporary buffer space... */

/* GUI */
	struct DiskObject              *Icon;
	struct Library                 *ThreadPoolBase;
	APTR                            ThreadPool;
};

/*********************************************************************************************/

#define	DEVICE_GADGET               1
#define	COMMENT_GADGET              2
#define	TEST_GADGET                 3
#define	SAVE_GADGET                 4
#define	STOP_GADGET                 5

/*********************************************************************************************/
/*
    This is the minimum timeval for test that can be extended/shorted automatically
    This number should not be set too low otherwise the test results will be
    inaccurate due to timer granularity.  (in seconds)
*/

#ifdef	SCSI_SPEED

#define	MIN_TEST_TIME               20

#else	/* SCSI_SPEED */

#define	MIN_TEST_TIME               8
#define	NUM_FILES                   200

#endif	/* SCSI_SPEED */

/*********************************************************************************************/
/*********************************************************************************************/
/*********************************************************************************************/

struct Catalog *MyCatalog;

STATIC ULONG getv(APTR obj, ULONG tag)
{
	ULONG x;
	GetAttr(tag, obj, &x);
	return x;
}

/****************************************************************************
	TimerInterrupt
****************************************************************************/

STATIC VOID TimerInterrupt(void)
{
	struct SpeedTimer *spt = (APTR)REG_A1;
	spt->TimeIsUp = TRUE;
}

STATIC struct EmulLibEntry TimerIntTrap = { TRAP_LIB, 0, &TimerInterrupt };

/*
                Start the timer...
*/

STATIC void Start_Timer(const struct DiskSpeed *global)
{
	struct SpeedTimer *spt = global->timer;
	struct Library *TimerBase = (struct Library *)(spt->tmReq.tr_node.io_Device);
	spt->TimeIsUp = FALSE;
	spt->tmReq.tr_time.tv_secs = global->Min_Time;
	spt->tmReq.tr_time.tv_micro = 0;
	SendIO(&spt->tmReq.tr_node);
	ReadCPUClock(&spt->StartTime);
}

/*
                Stop the timer...
*/

STATIC void Stop_Timer(struct SpeedTimer *spt)
{
	struct Library *TimerBase = (struct Library *)(spt->tmReq.tr_node.io_Device);
	UQUAD stoptime;
	DOUBLE ticks;

	ticks = (DOUBLE)ReadCPUClock(&stoptime);

	spt->FinalTime = ((DOUBLE)(stoptime - spt->StartTime)) / ticks;

	AbortIO(&spt->tmReq.tr_node);
	WaitIO(&spt->tmReq.tr_node);
}

STATIC void StartBenchmark(const struct DiskSpeed *global)
{
	Start_Timer(global);
}

STATIC void StopBenchmark(struct DiskSpeed *global)
{
	UQUAD cputime, ticks;

	Stop_Timer(global->timer);

	NewGetSystemAttrsA(&cputime, sizeof(cputime), SYSTEMINFOTYPE_LASTSECCPUTIME, NULL);
	NewGetSystemAttrsA(&ticks, sizeof(ticks), SYSTEMINFOTYPE_LASTSECTICKS, NULL);

	global->CPU_Usage = 100.0 - (DOUBLE)cputime / (DOUBLE)ticks * 100.0;
}

/*
                Free a SpeedTimer structure as best as possible.  Do all of the error checks
                here since this will also be called for partial timer initializations.
*/

STATIC void Free_Timer(struct SpeedTimer *spt)
{
	if (spt)
	{
		if (spt->Open)
			CloseDevice(&spt->tmReq.tr_node);

		FreeMem(spt,sizeof(struct SpeedTimer));
	}
}

/*
                Initialize a SpeedTimer structure.  It will return NULL if it did not work.
*/

STATIC struct SpeedTimer *Init_Timer(void)
{
    struct	SpeedTimer      *spt;

	if ((spt = AllocMem(sizeof(struct SpeedTimer),MEMF_ANY)))
	{
		spt->tmInt.is_Node.ln_Type = NT_UNKNOWN;
		spt->tmInt.is_Data = spt;
		spt->tmInt.is_Code = (APTR)&TimerIntTrap;

		spt->tmPort.mp_Flags = PA_SOFTINT;
		spt->tmPort.mp_SigTask = &spt->tmInt;
		NEWLIST(&spt->tmPort.mp_MsgList);

		spt->tmReq.tr_node.io_Message.mn_ReplyPort = &spt->tmPort;
		spt->tmReq.tr_node.io_Flags = 0;
		spt->tmReq.tr_node.io_Command = TR_ADDREQUEST;

		if (!OpenDevice(TIMERNAME, UNIT_VBLANK, &spt->tmReq.tr_node, 0))
		{
			spt->Open = TRUE;
		}
		else
		{
			Free_Timer(spt);
			spt = NULL;
		}
	}
	return(spt);
}

/*
                Now, for the routines that will pull in the lines and display them as
                needed in the display...
*/

STATIC void AddDisplayLine(const struct DiskSpeed *global, CONST_STRPTR line)
{
	ULONG size = strlen(line);

	if (diskspeedmui_intern.disk_main_win) /* add the line to the zune listview */
	{
		methodstack_push(diskspeedmui_intern.resultlv, 3, MUIM_List_InsertSingle, (IPTR)line, MUIV_List_Insert_Bottom);
		methodstack_pushsync(diskspeedmui_intern.resultlv, 2, MUIM_List_Jump, 0x7fffffff);
	}
	else if (global->Output)
	{
		Write(global->Output,(APTR)line,size);
		Write(global->Output,"\n",1);
	}
}

STATIC void AddDisplayLineFormat(const struct DiskSpeed *global, CONST_STRPTR fmt, ...)
{
	TEXT buf[512];
	va_list va;
	va_start(va, fmt);
	VNewRawDoFmt(fmt, NULL, buf, va);
	va_end(va);
	AddDisplayLine(global, buf);
}

/*...*/
STATIC LONG Abort_Testing;
STATIC LONG Check_Quit(struct DiskSpeed *global)
{
#if 0
	LONG worked = TRUE;
 
	if (SetSignal(0,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C) worked = FALSE;

	if (!worked) AddDisplayLine(global,GSI(MSG_USERINT));

	return(worked);
#else
	return Abort_Testing ? FALSE : TRUE;
#endif
}

/*
                Build a string and add it to the display of results.
                This routine knows how to take the timer results and the CPU
                results and format them into a string with the given header
                and the given unit of measure.
*/

STATIC VOID Display_Result(struct DiskSpeed *global, CONST_STRPTR Header,ULONG number, CONST_STRPTR Units)
{
	CONST_STRPTR format;
	ULONG clicks = 0;                     /* To figure out the number of clicks/second */

	number = ((DOUBLE)number / global->timer->FinalTime);

	if (number == 0)
	{
		number = 1;
		format = GSI(MSG_RESULT2_CPUAVAIL);
	}
	else
	{
		format = GSI(MSG_RESULT_CPUAVAIL);
	}

	clicks = global->CPU_Usage;

	if (clicks < 0)
		clicks = 0;
	if (clicks > 100)
		clicks = 100;

	global->CPU_Total += clicks;
	global->CPU_Count++;

	AddDisplayLineFormat(global, format, Header, number, Units, clicks);
}

STATIC VOID Display_Error(const struct DiskSpeed *global, CONST_STRPTR test)
{
	AddDisplayLineFormat(global, GSI(MSG_ERR_FAIL),test);
}

#ifdef	SCSI_SPEED

STATIC BOOL SpeedTest(struct DiskSpeed *global,ULONG size,ULONG offset)
{
    BOOL	            worked=TRUE;
    char	            *buffer;
    char	            *mem;		                            /* What we really allocated */
    char	            *type2;
    ULONG	            count;

	AddDisplayLine(global,"");

	type2="LONG";
	if (offset & 2) type2 = "WORD";
	if (offset & 1) type2 = "BYTE";

	/* Round to block sizes */
	size = (size+511) & (~511);

	if ((mem = AllocMem(size+offset, MEMF_ANY)))
	{
		/* Set up memory... */
		buffer = &(mem[offset]);

		AddDisplayLineFormat(global, GSI(MSG_TSTBUFF),size,type2);

		count = 0;

		StartBenchmark(global);

		while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
		{
			global->DiskIO.iotd_Req.io_Command = CMD_READ;
			global->DiskIO.iotd_Req.io_Flags = NULL;
			global->DiskIO.iotd_Req.io_Length = size;
			global->DiskIO.iotd_Req.io_Data = buffer;
			global->DiskIO.iotd_Req.io_Offset = count;

			DoIO((struct IORequest *)&global->DiskIO);
			count += global->DiskIO.iotd_Req.io_Actual;

			if (global->DiskIO.iotd_Req.io_Error)
			{
				worked = FALSE;
				Display_Error(global,GSI(MSG_ERR_RDEV));
			}
		}
		StopBenchmark(global);

		#if defined(SCSI_SPEED)
		if (worked) Display_Result(global,GSI(MSG_BYTES_READ_S),count,GSI(MSG_BYTE_UNITS));
		#else
		if (worked) Display_Result(global,GSI(MSG_BYTES_READ_N),count,GSI(MSG_BYTE_UNITS));
		#endif

		FreeMem(mem,size+offset);
	}
	else
	{
		AddDisplayLineFormat(global, GSI(MSG_SKIPMEM), size);
	}

	return(worked);
}

#else	/* SCSI_SPEED */

/*
                In order to keep the file create test fair, it must always do
                the same number of files.  The way filing systems work, many times
                the get slower as the number of files in a directory grow
*/

STATIC LONG CreateFileTest(struct DiskSpeed *global)
{
	TEXT filename[108];
    ULONG	            count;
    LONG	            worked = TRUE;

	StartBenchmark(global);

	for (count=0;(count<NUM_FILES) && (worked &= Check_Quit(global));count++)
	{
		BPTR file;

		NewRawDoFmt(FILE_STRING, NULL, filename, count);
		if ((file = Open(filename,MODE_NEWFILE))) Close(file);
		else
		{
			Display_Error(global,GSI(MSG_ERR_CFILE));
			worked = FALSE;
		}
	}

	StopBenchmark(global);

	if (worked) Display_Result(global,GSI(MSG_FILE_CREATE),NUM_FILES,GSI(MSG_FILE_UNITS));

	return(worked);
}

STATIC LONG OpenFileTest(struct DiskSpeed *global)
{
	TEXT filename[108];
    ULONG	            count = 0;
    LONG	            worked = TRUE;

	StartBenchmark(global);

	while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
	{
		BPTR file;

		NewRawDoFmt(FILE_STRING, NULL, filename, (count % NUM_FILES));
		count++;
		if ((file=Open(filename,MODE_OLDFILE))) Close(file);
		else
		{
			Display_Error(global,GSI(MSG_ERR_OFILE));
			worked = FALSE;
		}
	}

	StopBenchmark(global);

	if (worked) Display_Result(global,GSI(MSG_FILE_OPEN),count,GSI(MSG_FILE_UNITS));

	return(worked);
}

STATIC LONG ScanDirectoryTest(struct DiskSpeed *global)
{
	struct FileInfoBlock fib;
	BPTR lock = (BPTR)NULL;
    ULONG	            count = 0;
    LONG	            worked = TRUE;

	StartBenchmark(global);

	while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
	{
		if (lock)
		{
			if (!ExNext(lock, &fib)) lock = (BPTR)NULL;
			else count++;
		}
		else
		{
			CurrentDir(lock = CurrentDir((BPTR)NULL));
			if (Examine(lock, &fib)) count++;
			else
			{
				Display_Error(global,GSI(MSG_ERR_DIRSCAN));
				worked = FALSE;
			}
		}
	}

	StopBenchmark(global);

	if (worked) Display_Result(global,GSI(MSG_FILE_SCAN),count,GSI(MSG_FILE_UNITS));

	return(worked);
}

/*
                In order to keep the file delete test fair, it must always do
                the same number of files.  The way filing systems work, many times
                the get slower as the number of files in a directory grow
*/

STATIC LONG DeleteFileTest(struct DiskSpeed *global)
{
    ULONG	            count;
    LONG	            worked = TRUE;
    char	            *st_string = global->tmp1;

	StartBenchmark(global);

	for (count = 0;(count<NUM_FILES) && (worked &= Check_Quit(global));count++)
	{
		sprintf(st_string,FILE_STRING,count);
		if (!DeleteFile(st_string))
		{
			Display_Error(global,GSI(MSG_ERR_DFILE));
			worked=FALSE;
		}
	}

	StopBenchmark(global);

	if (worked) Display_Result(global,GSI(MSG_FILE_DELETE),NUM_FILES,GSI(MSG_FILE_UNITS));

	return(worked);
}

STATIC LONG SeekReadTest(struct DiskSpeed *global)
{
    BPTR	            file;
    ULONG	            size;
    ULONG	            count = count;
    LONG	            pos = 0;
    LONG	            buffer[16];
    LONG	            worked = FALSE;
    void	            *buf = buf;

	/* First we build a file by writing the ROM to disk... */
	if ((file = Open(FILE_STRING,MODE_NEWFILE)))
	{
		size = 0x40000;	/* Start by asking for 256K */
		while (size && (!(buf = AllocMem(size,MEMF_ANY)))) size = size>>1;

		if (buf)
		{
			worked = TRUE;
			/* Write a 256K file... */
			count = 0x40000/size;
			while ((count>0) && (worked&=Check_Quit(global)))
			{
				count--;
				if (size != Write(file,buf,size))
				{
					worked=FALSE;
					Display_Error(global,GSI(MSG_ERR_SEEK));
				}
			}
			FreeMem(buf,size);
		}
		else Display_Error(global,GSI(MSG_ERR_SEEK));

		StartBenchmark(global);

		while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
		{
			Seek(file,pos,OFFSET_BEGINNING);
			Read(file,buffer,64);
			count++;

			Seek(file,-(pos+64),OFFSET_END);
			Read(file,buffer,64);
			count++;

			Seek(file,-(pos+(size/3)),OFFSET_CURRENT);
			Read(file,buffer,64);
			count++;

			/* Come up with another position... */
			pos=(pos+(size/11)) % (size/3);
		}

		StopBenchmark(global);

		if (worked) Display_Result(global,GSI(MSG_SEEK_READ),count,GSI(MSG_SEEK_UNITS));

		Close(file);
		DeleteFile(FILE_STRING);
	}
	else Display_Error(global,GSI(MSG_ERR_SEEK));

	return(worked);
}

STATIC BOOL SpeedTest(struct DiskSpeed *global,ULONG size,ULONG offset)
{
	static const TEXT tmpfilename[] = "DiskSpeed RW Test File";

    BOOL	worked=TRUE;
    char	*buffer;
    char	*mem;		/* What we really allocated */
	CONST_STRPTR type2;
    ULONG	loop;
    ULONG	count;
    LONG	times;
	BPTR file = (BPTR)NULL;

	AddDisplayLine(global,"");

	type2="LONG";
	if (offset & 2) type2="WORD";
	if (offset & 1) type2="BYTE";

	if ((mem=AllocMem(size+offset,MEMF_ANY)))
	{
		/* Set up memory... */
		buffer=&(mem[offset]);

		for (loop=0;loop<size;loop++) buffer[loop]=(unsigned short )loop;

		AddDisplayLineFormat(global, GSI(MSG_TSTBUFF), size, type2);

		count = 0;
		times = 0;

		StartBenchmark(global);

		while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
		{
			if (times<1)
			{
				if (file) Close(file);
				DeleteFile(tmpfilename);
				if (!(file = Open(tmpfilename, MODE_NEWFILE)))
				{
					Display_Error(global,GSI(MSG_ERR_CFILE));
					worked = FALSE;
				}
				times = 0x40000/size;	/* Try to make file at least 256K size */
			}
			if (file)
			{
				if (size != Write(file,buffer,size))
				{
					Display_Error(global,GSI(MSG_ERR_CFILE));
					worked = FALSE;
				}
				else count += size;
				times--;
			}
		}

		StopBenchmark(global);

		if (worked) Display_Result(global,GSI(MSG_BYTES_CREATE),count,GSI(MSG_BYTE_UNITS));

		/* Fill out the file... */
		if (file) while ((worked &= Check_Quit(global)) && (times>0))
		{
			Write(file,buffer,size);
			times--;
		}

		if (file)
			Close(file);
		file = (BPTR)NULL;

		if (worked) if (!(file=Open(tmpfilename, MODE_OLDFILE)))
		{
			Display_Error(global,GSI(MSG_ERR_WFILE));
			worked=FALSE;
		}

		count=0;
		times=0;

		StartBenchmark(global);

		while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
		{
			if (times < 1)
			{
				Seek(file,0,OFFSET_BEGINNING);
				times = 0x40000/size;	/* Try to make file at least 256K size */
			}
			if (size != Write(file,buffer,size))
			{
				Display_Error(global,GSI(MSG_ERR_WFILE));
				worked = FALSE;
			}
			else count += size;
			times--;
		}

		StopBenchmark(global);

		if (worked) Display_Result(global,GSI(MSG_BYTES_WRITE),count,GSI(MSG_BYTE_UNITS));

		if (file) Close(file);
		file = (BPTR)NULL;

		if (worked) if (!(file = Open(tmpfilename, MODE_OLDFILE)))
		{
			Display_Error(global,GSI(MSG_ERR_RFILE));
			worked = FALSE;
		}

		count = 0;
		times = 0;

		StartBenchmark(global);

		while ((worked &= Check_Quit(global)) && !global->timer->TimeIsUp)
		{
			if (times < 1)
			{
				Seek(file,0,OFFSET_BEGINNING);
				times = 0x40000/size;	/* Try to make file at least 256K size */
			}
			if (size != Read(file,buffer,size))
			{
				Display_Error(global,GSI(MSG_ERR_RFILE));
				worked = FALSE;
			}
			else count += size;
			times--;
		}

		StopBenchmark(global);

		if (worked)
		{
			for (loop=0;loop<size;loop++) worked &= (buffer[loop] == (char)loop);
			if (!worked) AddDisplayLine(global,GSI(MSG_ERR_DATA));
		}

		#if defined(SCSI_SPEED)
		if (worked) Display_Result(global,GSI(MSG_BYTES_READ_S),count,GSI(MSG_BYTE_UNITS));
		#else
		if (worked) Display_Result(global,GSI(MSG_BYTES_READ_N),count,GSI(MSG_BYTE_UNITS));
		#endif

		if (file)
			Close(file);

		FreeMem(mem,size+offset);
		DeleteFile(tmpfilename);
	}
	else
	{
		AddDisplayLineFormat(global, GSI(MSG_SKIPMEM), size);
	}

	return(worked);
}

/*
                Clean up (remove) all of the files in the current directory...
*/

STATIC void CleanUpFiles(struct DiskSpeed *global)
{
	struct FileInfoBlock fib;
	BPTR lock;

	CurrentDir(lock = CurrentDir((BPTR)NULL));	/* Get current directory lock */

	while (lock)
	{
		if (Examine(lock, &fib))
		{
			if (ExNext(lock, &fib))
				DeleteFile(fib.fib_FileName);
			else
				lock = (BPTR)NULL;
		}
		else
		{
			lock = (BPTR)NULL;
		}
	}
}

#endif	/* SCSI_SPEED */

STATIC void DoTests(struct DiskSpeed *global)
{
    char                *st_string=global->tmp1;
    BOOL                working;
    ULONG               offset;
    short               size;

#ifndef	SCSI_SPEED
    char                *fstring;
    LONG                buffers;
#endif	/* !SCSI_SPEED */

	/*
	    Ok, so now we are ready to run...  Display the
	    test conditions...
	*/

	#if defined(__AROS__)
	AddDisplayLineFormat(global, "%s%s AROS: %s", GSI(MSG_PROCESSOR), global->CPU_Type, global->Exec_Ver);
	#else
	AddDisplayLineFormat(global, "%s%s MorphOS: %s", GSI(MSG_PROCESSOR), global->CPU_Type, global->Exec_Ver);
	#endif

	/*
	    Now, if we are in 2.0 OS, we can also find out the number of buffers
	    (maybe) on that device.  This is important.
	*/
#ifdef	SCSI_SPEED

	sprintf(st_string,"Device: %s",global->Device);

#else

	fstring = "Device: %s    Buffers: <information unavailable>";
	#ifndef __MORPHOS__
	if (DOSBase->dl_lib.lib_Version > 36L)
	#endif
	{
		/*
		    Ok, so we can now try to get a reading of the buffers
		    for the place we are about to test...

		    Note:  Since we are in the "CURRENTDIR" of the test disk,
		    we are using "" as the "device" to which to call AddBuffers()
		*/
		if ((buffers = AddBuffers("",0)) > 0) fstring="Device:  %s    Buffers: %ld";
	}

	sprintf(st_string,fstring,global->Device,buffers);

#endif	/* SCSI_SPEED */

	AddDisplayLine(global,st_string);

	if (global->Comments[0])
	{
		AddDisplayLineFormat(global, "%s%s", GSI(MSG_CMMNTS), global->Comments);
	}

	AddDisplayLine(global,"");

	global->CPU_Total = 0L;
	global->CPU_Count = 0L;

	working = Check_Quit(global);

#ifndef	SCSI_SPEED

	if (working && global->Test_DIR)
	{
		AddDisplayLine(global,GSI(MSG_TSTMANP));
		if (working) working = CreateFileTest(global);
		if (working) working = OpenFileTest(global);
		if (working) working = ScanDirectoryTest(global);
		if (working) working = DeleteFileTest(global);
	}

	if (working && global->Test_SEEK)
	{
		AddDisplayLine(global,"");
		if (working) working = SeekReadTest(global);
	}

#endif	/* !SCSI_SPEED */

	/*  Now for some of the more complex tests              */
	/*  result=SpeedTest(global,Buffer,offset,mem_type);    */
	if (global->Align_Types == 0)
		global->Align_Types = 4;

	for (offset = 4;offset>0;offset=offset >> 1) if (offset & global->Align_Types)
	{
		for (size = 0;size<4;size++) if (global->Test_Size[size])
		{
			if (working) working = SpeedTest(global,global->Test_Size[size],offset);
		}
	}

#ifndef	SCSI_SPEED

    CleanUpFiles(global);

#endif	/* !SCSI_SPEED */

	if (working)
	{
		AddDisplayLine(global,"");
		global->CPU_Total = (((global->CPU_Total << 1) / global->CPU_Count)+1) >> 1;
		AddDisplayLineFormat(global, GSI(MSG_AVGCPU), global->CPU_Total);
	}
}

#ifdef	SCSI_SPEED

STATIC void StartTest(struct DiskSpeed *global)
{
    APTR	            oldwindow;
    char	            *st_string;
    char	            *unit = NULL;
    ULONG	            scsi_unit;

	oldwindow = global->Me->pr_WindowPtr;
	global->Me->pr_WindowPtr = (APTR)(-1L);

	AddDisplayLine(global,COPYRIGHT);
	AddDisplayLine(global,STR_SEPERATOR);

	st_string = global->Device;
	while (*st_string)
	{
		if (*st_string == ':') unit = st_string;
		st_string++;
	}

	st_string = unit;

	if (unit)
	{
		*unit = '\0';
		unit++;
		scsi_unit = 0;
		while ((unit) && (*unit))
		{
			if ((*unit < '0') || (*unit >'9')) unit = NULL;
			else
			{
				scsi_unit *= 10;
				scsi_unit += *unit-'0';
				unit++;
			}
		}
	}

	if (unit)
	{
		memset(&(global->DiskIO),0,sizeof(struct IOExtTD));
		if ((global->DiskIO.iotd_Req.io_Message.mn_ReplyPort = CreateMsgPort()))
		{
			if (!OpenDevice(global->Device,scsi_unit,(struct IORequest *)&global->DiskIO,NULL))
			{
				unit = NULL;
				*st_string = ':';
				DoTests(global);
				CloseDevice((struct IORequest *)&global->DiskIO);
			}
			DeleteMsgPort(global->DiskIO.iotd_Req.io_Message.mn_ReplyPort);
		}
		if (unit) AddDisplayLine(global,GSI(MSG_ERR_NODEV));
	}
	else AddDisplayLine(global,GSI(MSG_ERR_BADSPEC));

	if (st_string) *st_string = ':';

	global->Me->pr_WindowPtr = oldwindow;
	methodstack_push(diskspeedmui_intern.disk_app, 2, MUIM_Application_ReturnID, RETURNID_BENCHMARK_FINISHED);
}

#else	/* SCSI_SPEED */

STATIC void StartTest(struct DiskSpeed *global)
{
	BPTR lock, newlock;
	APTR oldwindow;

	oldwindow = global->Me->pr_WindowPtr;
	global->Me->pr_WindowPtr = (APTR)(-1L);

	AddDisplayLine(global,COPYRIGHT);
	AddDisplayLine(global,STR_SEPERATOR);

	if ((lock = Lock(global->Device,ACCESS_READ)))
	{
		lock = CurrentDir(lock);

		DeleteFile(TEST_DIR);

		if ((newlock = CreateDir(TEST_DIR)))
		{
			UnLock(newlock);
			if ((newlock = Lock(TEST_DIR,ACCESS_READ)))
			{
				newlock = CurrentDir(newlock);

				/*
				    Now do all of the tests...
				*/

				DoTests(global);

				UnLock(CurrentDir(newlock));
			}
			else AddDisplayLine(global,GSI(MSG_ERR_BADDIR));

			DeleteFile(TEST_DIR);
		}
		else
		{
			AddDisplayLine(global,GSI(MSG_ERR_NODIR));
		}

		UnLock(CurrentDir(lock));
	}
	else AddDisplayLine(global,GSI(MSG_ERR_NOLOCK));

	global->Me->pr_WindowPtr = oldwindow;
	methodstack_push(diskspeedmui_intern.disk_app, 2, MUIM_Application_ReturnID, RETURNID_BENCHMARK_FINISHED);
}

#endif	/* SCSI_SPEED */

STATIC VOID SetVersionStrings(struct DiskSpeed *global)
{
	ULONG sysVer = ((struct ExecBase *)(SysBase))->LibNode.lib_Version;
	ULONG sofVer = ((struct ExecBase *)(SysBase))->SoftVer;

	#ifdef __PPC__
	global->CPU_Type = "PPC";
	#else
	global->CPU_Type = "i386";
	#endif

	NewRawDoFmt("%ld.%ld", NULL, global->Exec_Ver, sysVer, sofVer);
}

/*
                This routine closes down the GUI
*/

STATIC void Close_GUI(struct DiskSpeed *global)
{
	MUI_DisposeObject(diskspeedmui_intern.disk_app);
}

STATIC Object *MakeCheck (CONST_STRPTR label, ULONG id, ULONG checked)
{
    Object *obj = MUI_MakeObject(MUIO_Checkmark, (IPTR)label);
    if (obj)
    	SetAttrs(obj, MUIA_CycleChain, 1, MUIA_ObjectID, id, MUIA_Selected, checked, TAG_DONE);
    return obj;
}

static APTR MakeButtonCat(CONST_STRPTR string)
{
    APTR obj;

	if ((obj = MUI_MakeObject(MUIO_Button, string)))
        SetAttrs(obj, MUIA_CycleChain, 1, TAG_DONE);

    return obj;
}

STATIC APTR MakeNumString(size_t id)
{
	return StringObject,
		StringFrame,
		MUIA_CycleChain, 1,
		MUIA_ObjectID, id,
		MUIA_String_Accept, (IPTR)"0123456789",
		MUIA_String_Format, MUIV_String_Format_Right,
	End;
}

VOID OpenAppConfigWindow()
{
	set(diskspeedmui_intern.disk_config_win, MUIA_Window_Open, TRUE);
}

STATIC size_t PrepareGUI(struct DiskSpeed *global)
{
	struct Library *ThreadPoolBase = OpenLibrary("threadpool.library", 0);

	if (ThreadPoolBase)
	{
		static const size_t tags[] = { THREADPOOL_Name, (size_t)"DiskSpeed Benchmark", THREADPOOL_Priority, 0, TAG_DONE };

		global->ThreadPoolBase = ThreadPoolBase;
		global->ThreadPool = CreateThreadPoolTagList(10, (struct TagItem *)&tags);
	}

	return (size_t)global->ThreadPool;
}

/*
                This routine is used to open the GUI...
*/

STATIC APTR Open_GUI(struct DiskSpeed *global, CONST_STRPTR iconname, int load_config)
{
	global->Icon = GetDiskObject(iconname);

	D(bug("Open_GUI()\n"));

	if (PrepareGUI(global))
	{
		static TEXT wintitle[60];
		struct DiskSpeedMUI_Intern *mui = &diskspeedmui_intern;
		APTR btConfig, btSave, btUse, btCancel;
		APTR ConfigWin;

		NewRawDoFmt(GSI(MSG_MAIN_WINTITLE), NULL, wintitle, APPNAME, VERSION, REVISION);
#if defined(SCSI_SPEED)
		NewRawDoFmt(GSI(MSG_Welcome), NULL, global->tmp2, GSI(MSG_W_DEVICE));
#else
		NewRawDoFmt(GSI(MSG_Welcome), NULL, global->tmp2, GSI(MSG_W_FILESYS));
#endif /* SCSI_SPEED */

        diskspeedmui_intern.disk_app = NewObject(getappclass(), NULL,
				MUIA_Application_DiskObject, (IPTR)global->Icon,
				MUIA_Application_Title, (IPTR)APPNAME,
				MUIA_Application_Version, (IPTR)&VERSTAG[1],
				MUIA_Application_Copyright, (IPTR)" 1995-2004 The AROS Development Team, 2012-2013 The MorphOS Team",
				MUIA_Application_Author, (IPTR)"The AROS Development Team, The MorphOS Team",
				MUIA_Application_Description, GSI(MSG_DESCRIPTION),
				MUIA_Application_Base, (IPTR)APPNAME,

				SubWindow, diskspeedmui_intern.disk_main_win = WindowObject,
	            MUIA_Window_Title, wintitle,
	            MUIA_Window_ID, MAKE_ID('D','W','I','N'),
					MUIA_Window_Height, MUIV_Window_Height_Visible(50),
					MUIA_Window_Width, MUIV_Window_Width_Visible(60),
					WindowContents, VGroup,
		            Child, VGroup, /* Main Panel */
    				    Child, TextObject,
						    TextFrame,
						    MUIA_Text_Contents, global->tmp2,
					    End,
							Child, ColGroup(3),
                                MUIA_HorizWeight,100,
					            Child, Label(GSI(MSG_DEVICE)),
                                Child, diskspeedmui_intern.devicestr = StringObject,
						            StringFrame,
												MUIA_Weight, 10000,
                                    MUIA_CycleChain, TRUE,
                                    MUIA_String_MaxLen, 255,
                                    MUIA_String_Contents, global->Device,
					            End,
      		                Child, btConfig = MakeButtonCat(GSI(MSG_GAD_CONFIG)),
                                Child, Label(GSI(MSG_CMMNTS)),
                                Child, diskspeedmui_intern.commentstr = StringObject,
						            StringFrame,
												MUIA_Weight, 10000,
                                    MUIA_CycleChain, TRUE,
                                    MUIA_String_MaxLen, 255,
					            End,
									Child, RectangleObject, MUIA_Weight, 0, End,
                        End,
				        Child, ListviewObject,  /* Our results view */
						    MUIA_Listview_List, diskspeedmui_intern.resultlv = ListObject,
							    ReadListFrame,
							    MUIA_Listview_Input, FALSE,
							    MUIA_List_ConstructHook, MUIV_List_ConstructHook_String,
							    MUIA_List_DestructHook, MUIV_List_DestructHook_String,
						    End,
				        End,
                    End,
		            Child, HGroup, /* save/use/cancel button row */
		                MUIA_FixHeight, 1,
		                MUIA_Group_SameWidth, TRUE,
		                Child, mui->btBegin = MakeButtonCat(GSI(MSG_START_TEST)),
		                Child, mui->btAbort = MakeButtonCat(GSI(MSG_STOP_TEST)),
                        Child, HSpace(0),
		                Child, mui->btSaveResults = MakeButtonCat(GSI(MSG_SAVE_RESULTS)),
		            End,
	            End,
	        End,

				//NEXT WINDOW.. CONFIG
				SubWindow, diskspeedmui_intern.disk_config_win = ConfigWin = WindowObject,
					MUIA_Window_Title, (IPTR)GSI(MSG_CONFIG_WINTITLE),
					MUIA_Window_ID, MAKE_ID('C','W','I','N'),
					WindowContents, VGroup,
						Child, HGroup,
							MUIA_Background, MUII_GroupBack,
							GroupFrameT(GSI(MSG_MISC_SETTINGS)),
							#ifndef SCSI_SPEED
							Child, ColGroup(2),
								Child, diskspeedmui_config.dir_checkmark = MakeCheck(GSI(MSG_DIRSCAN_SETTING), MAKE_ID('C','D','I','R'), TRUE),
								Child, (IPTR) LLabel(GSI(MSG_DIRSCAN_SETTING)),
								Child, diskspeedmui_config.seek_checkmark = MakeCheck(GSI(MSG_SEEK_SETTING), MAKE_ID('S','E','E','K'), TRUE),
								Child, (IPTR) LLabel(GSI(MSG_SEEK_SETTING)),
							End, 
							#endif
							Child, HSpace(0),
							Child, VGroup,
								Child, (IPTR) CLabel(GSI(MSG_MINTIME_SETTING)),
								Child, diskspeedmui_config.mintime = MakeNumString(MAKE_ID('M','I','N','T')),
							End, 
						End, /* Misc Options */

						Child, VGroup,
							GroupFrameT(GSI(MSG_RWBUFS)),
							Child, ColGroup(2),
								Child, LLabel(GSI(MSG_BUF1TXT)),
								Child, diskspeedmui_config.buffer1 = MakeNumString(MAKE_ID('B','U','F','1')),
								Child, LLabel(GSI(MSG_BUF2TXT)),
								Child, diskspeedmui_config.buffer2 = MakeNumString(MAKE_ID('B','U','F','2')),
								Child, LLabel(GSI(MSG_BUF3TXT)),
								Child, diskspeedmui_config.buffer3 = MakeNumString(MAKE_ID('B','U','F','3')),
								Child, LLabel(GSI(MSG_BUF4TXT)),
								Child, diskspeedmui_config.buffer4 = MakeNumString(MAKE_ID('B','U','F','4')),
							End,
						End, /* Buffer Sizes */

						Child, VGroup,
							GroupFrameT(GSI(MSG_RWBUFS)),
							Child, ColGroup(3),
								Child, diskspeedmui_config.long_checkmark = MakeCheck(NULL, MAKE_ID('L','O','N','G'), TRUE),
								Child, (IPTR) LLabel("LONG"),
								Child, HSpace(0),
								Child, diskspeedmui_config.word_checkmark = MakeCheck(NULL, MAKE_ID('W','O','R','D'), TRUE),
								Child, (IPTR) LLabel("WORD"),
								Child, HSpace(0),
								Child, diskspeedmui_config.byte_checkmark = MakeCheck(NULL, MAKE_ID('B','Y','T','E'), TRUE),
								Child, (IPTR) LLabel("BYTE"),
								Child, HSpace(0),
							End, 
						End, /* Buffer Types */

						Child, HGroup,
							Child, btSave = MakeButtonCat(GSI(MSG_GAD_SAVE)),
							Child, HSpace(30),
							Child, btUse = MakeButtonCat(GSI(MSG_GAD_USE)),
							Child, HSpace(30),
							Child, btCancel = MakeButtonCat(GSI(MSG_GAD_CANCEL)),
						End,
					End,
				End,
			End;        

        if (diskspeedmui_intern.disk_app)
        {
            if (load_config)
            {
                /* Load settings
                **
                ** Ditched config items stored to icon tooltypes. However DRIVE and COMMENT
                ** options are recognized still.
                */

                DoMethod(diskspeedmui_intern.disk_app, MUIM_Application_Load, MUIV_Application_Load_ENV);

                if (global->Icon)
                {
                    CONST_STRPTR s;

                    if ((s = FindToolType(global->Icon->do_ToolTypes, "DRIVE")))
                    {
                        stccpy(global->Device, s, sizeof(global->Device));
                    }

                    if ((s = FindToolType(global->Icon->do_ToolTypes, "COMMENT")))
                    {
                        stccpy(global->Comments, s, sizeof(global->Comments));
                    }
                }
            }

            /* Setup notify and other stuff */

/*********/
            DoMethod(diskspeedmui_intern.disk_main_win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);

            DoMethod(btConfig, MUIM_Notify, MUIA_Pressed, FALSE, (IPTR)diskspeedmui_intern.disk_config_win, 3, MUIM_Set, MUIA_Window_Open, TRUE);

            //DoMethod(diskspeedmui_intern.stopobj, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, RETURNID_STOP);
            DoMethod(mui->btSaveResults, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, RETURNID_SAVE);
            DoMethod(mui->btBegin, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, RETURNID_BEGIN);
            DoMethod(mui->btAbort, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 2, MUIM_Application_ReturnID, RETURNID_STOP);
/**/
            DoMethod(diskspeedmui_intern.disk_config_win, MUIM_Notify, MUIA_Window_CloseRequest, TRUE, diskspeedmui_intern.disk_config_win, 3, MUIM_Set, MUIA_Window_Open, FALSE);
            DoMethod(btSave, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 3, MM_Application_SaveConfig, 1, ConfigWin);
            DoMethod(btUse, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Application, 3, MM_Application_SaveConfig, 0, ConfigWin);
            DoMethod(btCancel, MUIM_Notify, MUIA_Pressed, FALSE, MUIV_Notify_Window, 3, MUIM_Set, MUIA_Window_Open, FALSE);

/**********/
            //set(btConfig, MUIA_HorizWeight, 0); /* Set up the weight for the image ( so it takes up least space..) */ 
/*********/
            set(diskspeedmui_intern.commentstr, MUIA_String_Contents, global->Comments);
            set(diskspeedmui_intern.devicestr, MUIA_String_Contents, global->Device);

				set(mui->btSaveResults, MUIA_Disabled, TRUE);
				set(mui->btAbort, MUIA_Disabled, TRUE);

            // Set some reasonable values if saved values are crap

            if (!getv(diskspeedmui_config.mintime, MUIA_String_Integer))
                set(diskspeedmui_config.mintime, MUIA_String_Integer, global->Min_Time);
            if (!getv(diskspeedmui_config.buffer1, MUIA_String_Integer))
                set(diskspeedmui_config.buffer1, MUIA_String_Integer, global->Test_Size[0]);
            if (!getv(diskspeedmui_config.buffer2, MUIA_String_Integer))
                set(diskspeedmui_config.buffer2, MUIA_String_Integer, global->Test_Size[1]);
            if (!getv(diskspeedmui_config.buffer3, MUIA_String_Integer))
                set(diskspeedmui_config.buffer3, MUIA_String_Integer, global->Test_Size[2]);
            if (!getv(diskspeedmui_config.buffer4, MUIA_String_Integer))
                set(diskspeedmui_config.buffer4, MUIA_String_Integer, global->Test_Size[3]);

            #if 0
                #ifdef	SCSI_SPEED			
                set( diskspeedmui_config.dir_checkmark, MUIA_Disabled, TRUE);
                set( diskspeedmui_config.seek_checkmark, MUIA_Disabled, TRUE);
                set( diskspeedmui_config.long_checkmark, MUIA_Selected, TRUE);
                set( diskspeedmui_config.long_checkmark, MUIA_Disabled, TRUE);
                set( diskspeedmui_config.word_checkmark, MUIA_Disabled, TRUE);
                set( diskspeedmui_config.byte_checkmark, MUIA_Disabled, TRUE);
                #else	/* SCSI_SPEED */
                if (global->Test_DIR) nnset( diskspeedmui_config.dir_checkmark, MUIA_Selected,TRUE);
                if (global->Test_SEEK) nnset( diskspeedmui_config.seek_checkmark, MUIA_Selected,TRUE);
                if (global->Align_Types & 4) nnset( diskspeedmui_config.long_checkmark, MUIA_Selected, TRUE);
                if (global->Align_Types & 2) nnset( diskspeedmui_config.word_checkmark, MUIA_Selected, TRUE);
                if (global->Align_Types & 1) nnset( diskspeedmui_config.byte_checkmark, MUIA_Selected, TRUE);
                #endif	/* SCSI_SPEED */
            #endif
        }
    }

    return(diskspeedmui_intern.disk_app);
}

/*
                This routine will append to the end of a file the results currently in memory...
*/
STATIC void Save_Results(const struct DiskSpeedMUI_Intern *mui)
{
	BPTR fh;

	if ((fh = Open(RESULTS_FILE,MODE_OLDFILE)))
	{
		Seek(fh,0,OFFSET_END);
		Write(fh,"\n\n\n",2);
	}
	else fh = Open(RESULTS_FILE,MODE_NEWFILE);

	if (fh)
	{
		size_t pos = 0;

		for (;;)
		{
			STRPTR entry;

			DoMethod(mui->resultlv, MUIM_List_GetEntry, pos, &entry);

			if (!entry)
				break;

			VFPrintf(fh, "%s\n", &entry);

			pos++;
		}

		Close(fh);
	}
}

/*
                This routine will retrieve the config settings from the MUI_Objects
 */
STATIC void GET_Config(struct DiskSpeed *global)
{
	ULONG tmp;

	/*
		Ok, now get the information back out...
	*/

    get(diskspeedmui_config.mintime, MUIA_String_Integer, &global->Min_Time);

    get(diskspeedmui_config.buffer1, MUIA_String_Integer, &global->Test_Size[0]);
    get(diskspeedmui_config.buffer2, MUIA_String_Integer, &global->Test_Size[1]);
    get(diskspeedmui_config.buffer3, MUIA_String_Integer, &global->Test_Size[2]);
    get(diskspeedmui_config.buffer4, MUIA_String_Integer, &global->Test_Size[3]);

    #ifndef SCSI_SPEED
    global->Test_DIR=0;
    get( diskspeedmui_config.dir_checkmark, MUIA_Selected, &tmp);
    if (tmp) global->Test_DIR=1;

    global->Test_SEEK=0;
    get( diskspeedmui_config.seek_checkmark, MUIA_Selected, &tmp);
    if (tmp) global->Test_SEEK=1;
    #endif

    global->Align_Types &= (~4);
    get( diskspeedmui_config.long_checkmark, MUIA_Selected,&tmp);
    if (tmp) global->Align_Types |= 1<<2;

    global->Align_Types &= (~2);
    get( diskspeedmui_config.word_checkmark, MUIA_Selected,&tmp);
    if (tmp) global->Align_Types |= 1<<1;
    
    global->Align_Types &= (~1);
    get( diskspeedmui_config.byte_checkmark, MUIA_Selected,&tmp);
    if (tmp) global->Align_Types |= 1;

    set (diskspeedmui_intern.disk_config_win, MUIA_Window_Open, FALSE);
}

/*
    DRIVE/K	- select drive  (Default is current directory or scsi.device:6)
    COMMENT/K	- set comment string
    ALL/S	- select all tests
    DIR/S	- setect DIR tests
    SEEK/S	- select SEEK tests
    LONG/S	- select LONG aligned tests
    WORD/S	- select WORD aligned tests
    BYTE/S	- select BYTE aligned tests
    BUF1/K/N	- select buffer size 1
    BUF2/K/N	- select buffer size 2
    BUF3/K/N	- select buffer size 3
    BUF4/K/N	- select buffer size 4
    MINTIME/K/N	- select the minimum test tmVal (default=8) in seconds
    WINDOW/S	- use the GUI even though started from the CLI
 */

/*
            do the command line parsing here...
 */
STATIC LONG ParseArg(struct DiskSpeed *global, IPTR *arglist)
{
	LONG window = FALSE, working = TRUE;
	int i;

	if (arglist[ARG_DRIVE])
	{
		stccpy(global->Device, (const char *)arglist[ARG_DRIVE], sizeof(global->Device));
	}

	if (arglist[ARG_COMMENT])
	{
		stccpy(global->Comments, (const char *)arglist[ARG_COMMENT], sizeof(global->Comments));
	}

	if (arglist[ARG_ALL])
	{
		#if !defined(SCSI_SPEED)
		global->Test_DIR = TRUE;
		global->Test_SEEK = TRUE;
		#endif
		global->Align_Types = 4|2|1;
	}

	for (i = ARG_BUF1; i <= ARG_BUF4; i++)
	{
		if (arglist[i])
		{
			size_t *ptr = (size_t *)arglist[i];
			global->Test_Size[i - ARG_BUF1] = *ptr;
		}
	}

	if (arglist[ARG_MINTIME])
	{
		size_t *ptr = (size_t *)arglist[ARG_MINTIME];
		global->Min_Time = *ptr;
	}

	#if !defined(SCSI_SPEED)
	if (arglist[ARG_DIR])
		global->Test_DIR = TRUE;
	if (arglist[ARG_SEEK])
		global->Test_SEEK = TRUE;
	#endif

	if (arglist[ARG_LONG])
		global->Align_Types |= 4;
	if (arglist[ARG_WORD])
		global->Align_Types |= 2;
	if (arglist[ARG_BYTE])
		global->Align_Types |= 1;

	if (arglist[ARG_WINDOW])
		window = TRUE;

	if (global->Min_Time < 1)
		global->Min_Time = 1;

	D(bug("ParseArg() loop().. FINISHED\n"));

	if (global->Min_Time < 1) global->Min_Time = 1;

    D(bug("ParseArg() - preparing to open MUI window\n"));

	if (window)
		working = (LONG)Open_GUI(global, NULL, 0);

	return(working);
}

STATIC void HandleAll(struct DiskSpeed *global)
{
	struct DiskSpeedMUI_Intern *mui = &diskspeedmui_intern;
	struct Library *ThreadPoolBase = global->ThreadPoolBase;
	ULONG sigs = 0, benchmarking = 0, running = 1;

	set(mui->disk_main_win, MUIA_Window_Open, TRUE);

	do    
	{
		LONG returnid = (LONG) DoMethod(mui->disk_app, MUIM_Application_NewInput, (IPTR) &sigs);

		switch (returnid)
		{
			case MUIV_Application_ReturnID_Quit:
				running = 0;
				break;

			case RETURNID_BEGIN:
			{
            ULONG str1, str2;

            GET_Config(global);

            str1 = getv(mui->commentstr, MUIA_String_Contents);
            str2 = getv(mui->devicestr, MUIA_String_Contents);

            strcpy(global->Comments, (char *)str1);
            strcpy(global->Device, (char *)str2);

				DoMethod(mui->resultlv, MUIM_List_Clear);

				set(mui->btBegin, MUIA_Disabled, TRUE);
				set(mui->btSaveResults, MUIA_Disabled, TRUE);
				set(mui->btAbort, MUIA_Disabled, FALSE);

				Abort_Testing = FALSE;

				if (QueueWorkItem(global->ThreadPool, (THREADFUNC)StartTest, global) != WORKITEM_INVALID)
					benchmarking = 1;

            break;
			}

			case RETURNID_STOP:
				Abort_Testing = TRUE;
            break;

			case RETURNID_SAVE:
				Save_Results(mui);
				break;

			case RETURNID_BENCHMARK_FINISHED:
				benchmarking = 0;
				set(mui->btBegin, MUIA_Disabled, FALSE);
				set(mui->btAbort, MUIA_Disabled, TRUE);

				if (getv(mui->resultlv, MUIA_List_Entries) > 0)
					set(mui->btSaveResults, MUIA_Disabled, FALSE);
				break;
		}
    	
		if (sigs)
		{
			sigs = Wait(sigs | SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_E);

			if (sigs & SIGBREAKF_CTRL_E)
				methodstack_check();

			if (sigs & SIGBREAKF_CTRL_C)
				running = 0;
	    }
	}
	while (benchmarking || running);
}

/*
                This routine is called when we want to run from a GUI
                Normally, it is only called when started from Workbench
                or when the CLI WINDOW option is given...
 */
STATIC void FromWanderer(struct DiskSpeed *global, struct WBStartup *startup)
{
	struct WBArg *wbarg = startup->sm_ArgList;

	if (wbarg && startup->sm_NumArgs)
	{
		BPTR lock;

		/*
		    Check if we were started as a project and
		    use that icon insted...
		*/
		if ((startup->sm_NumArgs) > 1) wbarg++;

		lock = CurrentDir(wbarg->wa_Lock);

		D(bug("FromWanderer() - launching GUI\n"));

		if (Open_GUI(global, wbarg->wa_Name, 1))
		{
			HandleAll(global);
		}

		CurrentDir(lock);
	}

	D(bug("FromWanderer() Finished..\n"));
}

STATIC ULONG ExceptCode()
{
	Abort_Testing = TRUE;
	return SIGBREAKF_CTRL_C;
}

STATIC struct EmulLibEntry ExceptTrap = { TRAP_LIB, 0, (void(*)())&ExceptCode };

/*
                This is the CLI starting point.  We do the command line parsing here...
 */
STATIC void FromCLI(struct DiskSpeed *global, IPTR *arglist)
{
	if (ParseArg(global, arglist))
	{
		if ( diskspeedmui_intern.disk_app )
		{
			HandleAll(global);
		}
		else
		{
			struct Task *task = SysBase->ThisTask;
			APTR OldExceptCode = task->tc_ExceptCode;

			task->tc_ExceptCode = (APTR)&ExceptTrap;
			SetExcept(SIGBREAKF_CTRL_C, SIGBREAKF_CTRL_C);

			StartTest(global);

			SetExcept(0, SIGBREAKF_CTRL_C);
			task->tc_ExceptCode = OldExceptCode;
		}
	}
	else
	{
		AddDisplayLine(global, CMDARG);
	}
}

STATIC VOID DeleteDiskSpeedCtx(struct DiskSpeed *global)
{
	if (global->ThreadPoolBase)
	{
		struct Library *ThreadPoolBase = global->ThreadPoolBase;

		if (global->ThreadPool)
			DeleteThreadPool(global->ThreadPool);

		CloseLibrary(ThreadPoolBase);
	}

	if (global->Icon)
		FreeDiskObject(global->Icon);

	Free_Timer(global->timer);

	FreeMem(global, sizeof(*global));
}

STATIC APTR CreateDiskSpeedCtx()
{
	struct DiskSpeed *global = AllocMem(sizeof(struct DiskSpeed), MEMF_CLEAR);

	if (global)
	{
		BOOL success = FALSE;

		global->Me = (struct Process *)SysBase->ThisTask;

		/* Standard MinTime */
		global->Min_Time=MIN_TEST_TIME;

		/* Standard sizes */
		global->Test_Size[0] = 0x200;
		global->Test_Size[1] = 0x1000;
		global->Test_Size[2] = 0x8000;
		global->Test_Size[3] = 0x40000;

		if ((global->timer = Init_Timer()))
		{
			#ifdef	SCSI_SPEED
			strcpy(global->Device, "ide.device:0");
			#else
			BPTR lock;

			if ((lock = Lock("", ACCESS_READ)))
			{
				char *p;

				NameFromLock(lock, global->Device, sizeof(global->Device));
				UnLock(lock);

				if ((p = strchr(global->Device, ':')))
				{
					p[1] = '\0';
				}
			}
			#endif	/* SCSI_SPEED */

			success = TRUE;
		}

		if (!success)
		{
			DeleteDiskSpeedCtx(global);
			global = NULL;
		}
	}

	return global;
}

int start(struct WBStartup *wbmsg, IPTR *arglist)
{
	struct DiskSpeed *global;

	if ((global = CreateDiskSpeedCtx()))
	{
		MyCatalog = OpenCatalogA(NULL, "DiskSpeed.catalog", NULL);

		SetVersionStrings(global);

		/*
			Now either set up Window or Output
			depending on where we were started...

			If we can not get Output, we set up the window...
		*/

		if (create_appclass())
		{
			methodstack_init();

			if ((arglist) && ((global->Output = Output())))
				FromCLI(global, arglist);
			else
				FromWanderer(global, wbmsg);

			Close_GUI(global);
			methodstack_cleanup();
			delete_appclass();
		}

		CloseCatalog(MyCatalog);
		DeleteDiskSpeedCtx(global);
	}

	return 0;
}
